package com.eu.miscedautils.mygentbvhdl;

import java.io.*;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.text.DateFormat;
import java.util.Date;

import com.eu.miscedautils.CommonUtils.LogMessages;
import com.eu.miscedautils.vhdlparser.FVhPBase;
import com.eu.miscedautils.vhdlparser.FVhPConstraint;
import com.eu.miscedautils.vhdlparser.FVhPEntity;
import com.eu.miscedautils.vhdlparser.FVhPExprBase;
import com.eu.miscedautils.vhdlparser.FVhPExprInteger;
import com.eu.miscedautils.vhdlparser.FVhPInterfaceElement;
import com.eu.miscedautils.vhdlparser.FVhPLibrary;
import com.eu.miscedautils.vhdlparser.FVhPPort;
import com.eu.miscedautils.vhdlparser.FVhPRange;
import com.eu.miscedautils.vhdlparser.FVhPSelectedName;
import com.eu.miscedautils.vhdlparser.FVhPSubTypeIndication;
import com.eu.miscedautils.vhdlparser.FVhPExprBase.ExprType;
import com.eu.miscedautils.vhdlparser.FVhPPort.Direction;

// CONV_INTEGER ==>  IEEE.STD_LOGIC_SIGNED.all
// TO_INTEGER  ==> IEEE.NUMERIC_STD/BIT
// CONV_ all types from ST_LOGIC : std_logic_arith 

public class FVhPGenTB {
	public enum PrefinedDataType {
		BIT, 
		BIT_VECTOR, 
		std_ulogic, 
		std_ulogic_vector, 
		std_logic, 
		std_logic_vector, 
		SIGNED, 
		UNSIGNED, 
		BOOLEAN, 
		CHARACTER, 
		STRING, 
		INTEGER, 
		NATURAL, 
		POSITIVE, 
		REAL, 
		SEVERITY_LEVEL, 
		TIME
	};

	private String logPrefix;
	public static Integer errorCount = 0;
	public static boolean isVerbose = false;
	private Integer inputWidth;
	private Integer tabCount;
	private FVhPEntity dut;

	private FileOutputStream outstream;
	private DataOutputStream outdata;
	private BufferedWriter output;
	private String randVarName;
	private String clkName;
	private String rstName;
	String strBuff;

	public FVhPGenTB(FVhPEntity top, String outFile, String clk, String rst) {
		top.elaborate();
		this.dut = top;
		this.clkName = clk;
		this.rstName = rst;
		this.tabCount = 0;
		strBuff = "";
		this.randVarName = new String("TB_indata_array");
		try {
			File file = new File(outFile);
			output = new BufferedWriter(new FileWriter(file));
		} catch (IOException e) {
			e.printStackTrace();
		}
		this.logPrefix = LogMessages.instance().getToolprefix();
	}

	public String getLogPrefix() {
		return logPrefix;
	}

	public void setLogPrefix(String logPrefix) {
		this.logPrefix = logPrefix;
	}

	public void printTB() {
		strBuff += printHeader();
		strBuff += printConversionFunctionPackage();
		strBuff += printConversionFunctionPackageBody();
		strBuff += "library STD;\r\n";
		strBuff += "library ieee;\r\n";
		strBuff += "use std.textio.all;\r\n";
		strBuff += "use ieee.math_real.all; -- for UNIFORM, TRUNC\r\n";
		strBuff += "use ieee.numeric_std.all; -- for TO_UNSIGNED\r\n";
		strBuff += "use ieee.std_logic_1164.all;\r\n";
		strBuff += "use ieee.std_logic_textio.all;\r\n";
		strBuff += "use work.TB_conv_pack.all;\r\n";

		// printLibraryAndUseClauses();

		String tb_name = "TB_" + dut.getName() + "_tb";
		strBuff += "\r\n\r\n";
		strBuff += "entity " + tb_name + " is\r\n";
		strBuff += "end entity " + tb_name + ";\r\n\r\n";
		strBuff += "architecture arch of " + tb_name + " is\r\n";
		increaseIndent();
		declareRandomVarStuff();
		declareSignals();
		decreaseIndent();
		strBuff += "begin\r\n";
		increaseIndent();
		assignRandomValues();
		instantiateDUT();
		declareProcessForRandomStimui();
		declareProcessForPrintingValueChanges();

		// declareClkRst();
		// initClkAndChangeVal();
		// initRstAndChangeValue();
		// changeInputPattern();
		// initFinish();
		decreaseIndent();
		indent();
		strBuff += "end;\r\n";

		try {
			output.write(strBuff);
			output.close();
		} catch (IOException e) {
			e.printStackTrace();
		}
		return;
	}

	void printLibraryAndUseClauses() {
		LinkedList<FVhPLibrary> libs = dut.getLibraries();

		if ( null != libs )
		{
			Iterator<FVhPLibrary> itr = libs.iterator();
			while (itr.hasNext()) 
			{
				FVhPLibrary lib = itr.next();
				if ( null == lib )
				{
					continue;
				}
				indent();
				strBuff += "library " + lib.print() + ";\r\n";
			}
		}
		LinkedList<FVhPSelectedName> useclauses = dut.getUseClauses();
		if ( null != useclauses )
		{
			Iterator<FVhPSelectedName> selNameItr = useclauses.iterator();
			while (selNameItr.hasNext()) 
			{
				FVhPSelectedName selname  = selNameItr.next();
				if ( null == selname ){
					continue;
				}
				indent();
				strBuff += "use " + selname.print() + ";\r\n";
			}
		}
		return;
	}

	public void declareRandomVarStuff() {
		Integer width = getTotalInputPortWidth();
		if (0 == width && null == clkName && null == rstName) {
			LogMessages.instance().print(  100, LogMessages.MSG_SEVERITY.ERROR, "No input port found in the top module ...");
			return;
		}
		// Declare a line as below-
		// signal random_input : std_logic_vector( 0 to 3 );
		String str = "signal " + randVarName + " : std_logic_vector( 0 to ";
		str += width - 1;
		str += ");\r\n ";
		indent();
		strBuff += str;
		// constant TB_CONST_ALL_1 : UNSIGNED ( 0 to 3 ) := ( others => '1' );
		str = "constant TB_CONST_ALL_1 : UNSIGNED ( 0 to ";
		str += width - 1;
		str += ") := ( others => '1' );\r\n ";
		indent();
		strBuff += str;
		// constant TB_MAX_LIMIT : integer := TO_INTEGER(TB_CONST_ALL_1);
		str = "constant TB_MAX_LIMIT : integer := TO_INTEGER(TB_CONST_ALL_1);\r\n";
		indent();
		strBuff += str;
		return;
	}

	void declareSignals() {
		List<FVhPPort> ports = dut.getPorts();
		Iterator itr = ports.iterator();
		while (itr.hasNext()) {
			FVhPPort port = (FVhPPort) itr.next();
			FVhPInterfaceElement intf_ele = port.getIntfEle();
			FVhPSubTypeIndication sind = intf_ele.getType();
			String str = "signal TB_" + intf_ele.getName();
			str += " : " + sind.print();
			str += ";\r\n";
			indent();
			strBuff += str;
		}
		return;
	}

	void instantiateDUT() {
		indent();
		strBuff += "inst : ";
		strBuff += "entity work." + dut.getName();
		strBuff += " port map \r\n";
		increaseIndent();
		increaseIndent();
		indent();
		strBuff += "(";
		List<FVhPPort> ports = dut.getPorts();
		Iterator itr = ports.iterator();
		Integer currIndex = 0;
		boolean prntComma = false;

		while (itr.hasNext()) {
			FVhPPort port = (FVhPPort) itr.next();
			FVhPInterfaceElement intf_ele = port.getIntfEle();

			if (true == prntComma) {
				strBuff += ", \r\n";
				indent();
			}
			prntComma = true;
			strBuff += intf_ele.getName();
			strBuff += " => ";

			strBuff += "TB_";
			strBuff += intf_ele.getName();
		}
		strBuff += "\r\n";
		indent();
		strBuff += ");\r\n";
		decreaseIndent();
		decreaseIndent();
		return;
	}

	void assignRandomValues() {

		List<FVhPPort> ports = dut.getPorts();
		Iterator itr = ports.iterator();
		Integer currIndex = 0;

		while (itr.hasNext()) {

			FVhPPort port = (FVhPPort) itr.next();
			FVhPInterfaceElement intf_ele = port.getIntfEle();
			FVhPSubTypeIndication sind = intf_ele.getType();
			Integer width = getSignalWidth(sind);

			if (FVhPPort.Direction.OUTPUT == port.getDirection()) {
				continue;
			}
			indent();
			strBuff += "TB_";
			strBuff += intf_ele.getName();
			strBuff += " <= ";
			strBuff += printConverterFromStdLogicVector(randVarName, sind,
					                                    currIndex, (currIndex + width - 1));
			strBuff += ";\r\n";
			currIndex += width;
		}

		return;
	}

	void declareProcessForRandomStimui() {
		indent();
		strBuff += "process\r\n";
		increaseIndent();
		indent();
		strBuff += "variable seed1, seed2: positive;\r\n";
		indent();
		strBuff += "variable rand: real;\r\n";
		indent();
		strBuff += "variable int_rand: integer;\r\n";
		decreaseIndent();
		indent();
		strBuff += "begin\r\n";
		increaseIndent();
		indent();
		strBuff += "loop\r\n";
		increaseIndent();
		indent();
		strBuff += "UNIFORM(seed1, seed2, rand);\r\n";
		indent();
		strBuff += "int_rand := INTEGER(TRUNC(rand*(real(TB_MAX_LIMIT))));\r\n";
		indent();
		strBuff += randVarName + " <= std_logic_vector(to_unsigned(int_rand,";
		strBuff += randVarName + "'LENGTH));\r\n";
		indent();
		strBuff += "wait for 5 ns;\r\n";
		decreaseIndent();
		indent();
		strBuff += "end loop;\r\n";

		decreaseIndent();
		indent();
		strBuff += "end process;\r\n";
		return;
	}

	void declareProcessForPrintingValueChanges() {
		indent();
		strBuff += "process( ";

		List<FVhPPort> ports = dut.getPorts();
		Iterator itr = ports.iterator();
		boolean printComma = false;
		while (itr.hasNext()) {
			FVhPPort port = (FVhPPort) itr.next();
			FVhPInterfaceElement intf_ele = port.getIntfEle();
			if (printComma)
				strBuff += ",";
			strBuff += "TB_" + intf_ele.getName();
			printComma = true;
		}
		strBuff += ")\r\n";
		increaseIndent();
		indent();
		strBuff += "variable var : line;\r\n";
		decreaseIndent();
		indent();
		strBuff += "begin\r\n";
		increaseIndent();
		indent();
		strBuff += "write(var,string'(\"Time = \"));\r\n";
		indent();
		strBuff += "write(var,now);\r\n";

		ports = dut.getPorts();
		itr = ports.iterator();
		while (itr.hasNext()) {
			FVhPPort port = (FVhPPort) itr.next();
			FVhPInterfaceElement intf_ele = port.getIntfEle();
			indent();
			strBuff += "write(var,string'(\" ";
			strBuff += intf_ele.getName();
			strBuff += " = \"));\r\n";
			indent();
			strBuff += "write(var,";
			strBuff += "TB_" + intf_ele.getName() + ");\r\n";
		}

		indent();
		strBuff += "writeline(OUTPUT,var);\r\n";
		decreaseIndent();
		indent();
		strBuff += "end process;\r\n";
		return;
	}

	void increaseIndent() {
		tabCount++;
	}

	void decreaseIndent() {
		tabCount--;
	}

	void indent() {
		for (int i = 0; i < tabCount; i++)
			strBuff += "\t";
	}

	Integer getSignalWidth(FVhPSubTypeIndication sind) {
		Integer width = 0;

		String typemark = sind.getTypeMark().print();
		if (false == isKnownType(typemark)) {
			LogMessages.instance().print( 101, LogMessages.MSG_SEVERITY.ERROR,  "Unsupported data type "
					+ typemark + " found in the interface ...");
			return 0;
		}
		
		LinkedList<FVhPBase> ranges = sind.getRanges();
		FVhPBase cons = null;
		FVhPRange range = null;
		if ( null != ranges )
		{
			if ( ranges.size() > 1 )
			{
				LogMessages.instance().print( 102,  LogMessages.MSG_SEVERITY.ERROR,  "Multidimensional array found at " + sind.getFile() + ":" + sind.getLine());
			}
			Iterator<FVhPBase> itr = ranges.iterator();
			while ( itr.hasNext() && null == cons )
			{
				cons = itr.next();
			}
		}
		if (null != cons) 
		{
			switch (cons.getObjType()) 
			{
			case CONSTRAINT:
			{
				FVhPConstraint mycons = (FVhPConstraint)cons;
				switch( mycons.getType())
				{
					case INDEXCONSTRAINT: 
					{
						LogMessages.instance().print( 103,  LogMessages.MSG_SEVERITY.ERROR, "Unsupported constraint type " + mycons.getType().toString() + " found at " +
								mycons.getFile() + ":" + mycons.getLine() + " ...");
						break;
					}
					case RANGECONSTRAINT: 
					{
						range = ((FVhPRange) cons);
						break;
					}
					default:
						LogMessages.instance().print( 104, LogMessages.MSG_SEVERITY.ERROR, "Unsupported constraint type " + mycons.getType().toString() + " found at " +
							mycons.getFile() + ":" + mycons.getLine() + " ...");
						break;
				}
				break;
			}
			case RANGE:
			{
				range = ((FVhPRange) cons);
				break;
			}
			default:
				LogMessages.instance().print( 105, LogMessages.MSG_SEVERITY.ERROR, "Unsupported constraint type " + cons.getObjType().toString() + " found at " +
						cons.getFile() + ":" + cons.getLine() + " ...");
				break;
			}
		}
		Integer rangeMax = 0;
		if (null != range) {
			range.elaborate();
			FVhPExprBase left = range.getLeft().evaluate();
			FVhPExprBase right = range.getRight().evaluate();
			if ( null == left  )
			{
				LogMessages.instance().print( 106, LogMessages.MSG_SEVERITY.ERROR, "File " + range.getFile() + ":" + range.getLine() + " left bound of the range '"+range.print() +"' is is not static ...");
				return 1;
			}
			if ( null == right )
			{
				LogMessages.instance().print( 107, LogMessages.MSG_SEVERITY.ERROR, "File " + range.getFile() + ":" + range.getLine() + " right bound of the range '"+range.print() +"' is is not static ...");
				return 1;
			}
			Integer lInt = 0, rInt = 0;
			if (FVhPExprBase.ExprType.CONSTINTEGER == left.getType()) 
			{
				lInt = ((FVhPExprInteger) left).getValue();
			}
			else
			{
				LogMessages.instance().print( 108, LogMessages.MSG_SEVERITY.ERROR, "Left bound of the range in " + range.getFile() + ":" + range.getLine() + " if of unsupported expr type " + left.getObjType().toString());
				
			}
			if (FVhPExprBase.ExprType.CONSTINTEGER == right.getType()) 
			{
				rInt = ((FVhPExprInteger) right).getValue();
			}
			else
			{
				LogMessages.instance().print( 109, LogMessages.MSG_SEVERITY.ERROR, "Right bound of the range in " + range.getFile() + ":" + range.getLine() + " if of unsupported expr type " + right.getObjType().toString());
				
			}
			width = (lInt - rInt);
			width = ((width < 0) ? (-1 * width) : width) + 1;
			rangeMax = (lInt > rInt) ? lInt : rInt;
		} 
		else 
		{
			width = 1;
		}
		if (typemark.equalsIgnoreCase("INTEGER")
				|| typemark.equalsIgnoreCase("NATURAL")
				|| typemark.equalsIgnoreCase("POSITIVE")) {
			width = 32;
			if (rangeMax > 0) {
				width = Integer.bitCount(rangeMax);
			}
		}
		return width;
	}

	Integer getTotalInputPortWidth() {
		List<FVhPPort> ports = dut.getPorts();
		Iterator itr = ports.iterator();

		Integer currIndex = 0;

		while (itr.hasNext()) {
			FVhPPort port = (FVhPPort) itr.next();
			if (FVhPPort.Direction.OUTPUT == port.getDirection()) {
				continue;
			}
			FVhPInterfaceElement intf_ele = port.getIntfEle();

			if ((null != clkName)
					&& intf_ele.getName().equalsIgnoreCase(clkName))
				continue;

			if ((null != rstName)
					&& intf_ele.getName().equalsIgnoreCase(rstName))
				continue;

			FVhPSubTypeIndication sind = intf_ele.getType();
			Integer width = getSignalWidth(sind);
			currIndex += width;
		}
		return currIndex;
	}

	boolean isKnownType(String type_mark) {
		boolean retVal = type_mark.equalsIgnoreCase("BIT")
				|| type_mark.equalsIgnoreCase("BIT_VECTOR")
				|| type_mark.equalsIgnoreCase("std_ulogic")
				|| type_mark.equalsIgnoreCase("std_ulogic_vector")
				|| type_mark.equalsIgnoreCase("std_logic")
				|| type_mark.equalsIgnoreCase("std_logic_vector")
				|| type_mark.equalsIgnoreCase("SIGNED")
				|| type_mark.equalsIgnoreCase("UNSIGNED")
				|| type_mark.equalsIgnoreCase("BOOLEAN")
				|| type_mark.equalsIgnoreCase("CHARACTER")
				|| type_mark.equalsIgnoreCase("STRING")
				|| type_mark.equalsIgnoreCase("INTEGER")
				|| type_mark.equalsIgnoreCase("NATURAL")
				|| type_mark.equalsIgnoreCase("POSITIVE");
		return retVal;
		// type_mark.equalsIgnoreCase("REAL") ||
		// type_mark.equalsIgnoreCase("SEVERITY_LEVEL") ||
		// type_mark.equalsIgnoreCase("TIME")
	}

	String printConverterFromStdLogicVector(String randomVarName,
			FVhPSubTypeIndication sind, Integer leftIndex, Integer rightIndex) {
		String type_mark = sind.getTypeMark().print();
		String str = new String("");
		if (type_mark.equalsIgnoreCase("BIT")) {
			str += "std_logic2bit( " + randomVarName + "(" + leftIndex + ") )";
		} else if (type_mark.equalsIgnoreCase("BOOLEAN")) {
			str += "std_logic2boolean( " + randomVarName + "(" + leftIndex
					+ ") )";
		} else if (type_mark.equalsIgnoreCase("CHARACTER")) {
			str += "std_logic2character( " + randomVarName + "(" + leftIndex
					+ ") )";
		} else if (type_mark.equalsIgnoreCase("INTEGER")) {
			str += "std_logic_vector2integer( " + randomVarName + "("
					+ leftIndex + " to " + rightIndex + ") )";
		} else if (type_mark.equalsIgnoreCase("NATURAL")) {
			str += "std_logic_vector2natural( " + randomVarName + "("
					+ leftIndex + " to " + rightIndex + ") )";
		} else if (type_mark.equalsIgnoreCase("positive")) {
			str += "std_logic_vector2positive( " + randomVarName + "("
					+ leftIndex + " to " + rightIndex + ") )";
		} else if (type_mark.equalsIgnoreCase("std_ulogic")) {
			str += "std_logic2std_ulogic( " + randomVarName + "(" + leftIndex
					+ ") )";
		} else if (type_mark.equalsIgnoreCase("std_logic")) {
			str += randomVarName + "(" + leftIndex + ")";
		} else if (type_mark.equalsIgnoreCase("std_ulogic_vector")) {
			str += "std_logic_vector2std_ulogic_vector( " + randomVarName + "("
					+ leftIndex + " to " + rightIndex + " ) )";
		} else if (type_mark.equalsIgnoreCase("std_logic_vector")) {
			str += randomVarName + "(" + leftIndex + " to " + rightIndex + " )";
		} else if (type_mark.equalsIgnoreCase("BIT_VECTOR")) {
			str += "std_logic_vector2bit_vector( " + randomVarName + "("
					+ leftIndex + " to " + rightIndex + " ) )";

		} else if (type_mark.equalsIgnoreCase("SIGNED")) {
			str += "std_logic_vector2signed( " + randomVarName + "("
					+ leftIndex + " to " + rightIndex + " ) )";

		} else if (type_mark.equalsIgnoreCase("UNSIGNED")) {
			str += "std_logic_vector2unsigned( " + randomVarName + "("
					+ leftIndex + " to " + rightIndex + " ) )";

		}
		else
		{
			LogMessages.instance().print( 110, LogMessages.MSG_SEVERITY.WARNING,  "You need to define a conversion function named  '" +
                    "std_logic_vector2" + type_mark + "' since you" );
            LogMessages.instance().print( 0, LogMessages.MSG_SEVERITY.NONE, 
                    		"               are using use-defined types for the port/generic of your DUT");
			str += "std_logic_vector2" + type_mark + "( " + randomVarName + "("
						+ leftIndex + " to " + rightIndex + " ) )";
		}
		return str;
	}

	String printConversionFunctionPackage() {
		String str = new String();
		str += "library ieee;\r\n";
		str += "use ieee.numeric_std.all;\r\n";
		str += "use ieee.std_logic_1164.all;\r\n";
		str += "\r\n\r\n";

		str += "package TB_conv_pack is\r\n";
		str += "    function std_logic_vector2unsigned(ARG : std_logic_vector ) return unsigned;\r\n";
		str += "    function std_logic_vector2integer(ARG : std_logic_vector ) return integer;\r\n";
		str += "    function std_logic_vector2bit_vector(ARG : std_logic_vector ) return bit_vector;\r\n";
		str += "    function std_logic_vector2natural(ARG : std_logic_vector ) return natural;\r\n";
		str += "    function std_logic_vector2std_ulogic_vector(ARG : std_logic_vector ) return std_ulogic_vector;\r\n";
		str += "    function std_logic_vector2signed(ARG : std_logic_vector ) return signed;\r\n";

		str += "    function std_logic2bit(ARG : std_logic ) return bit;\r\n";
		str += "    function std_logic2std_ulogic(ARG : std_logic ) return std_ulogic;\r\n";
		str += "    function std_logic2boolean(ARG : std_logic ) return boolean;\r\n";
		str += "    function std_logic2character(ARG : std_logic ) return character;\r\n";
		str += "end package TB_conv_pack;\r\n";
		str += "\r\n";
		return str;
	}

	String printConversionFunctionPackageBody() {
		String str = new String();
		str += "package body TB_conv_pack is\r\n";
		str += "function std_logic_vector2unsigned(ARG : std_logic_vector ) return unsigned is\r\n";
		str += "    variable var1 : unsigned(ARG'range);\r\n";
		str += "begin\r\n";
		str += "    for I in ARG'range loop\r\n";
		str += "        var1(I) := ARG(I);\r\n";
		str += "    end loop;\r\n";
		str += "    return var1;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";

		str += "function std_logic_vector2signed(ARG : std_logic_vector ) return signed is\r\n";
		str += "    variable var1 : signed(ARG'range);\r\n";
		str += "begin\r\n";
		str += "    for I in ARG'range loop\r\n";
		str += "        var1(I) := ARG(I);\r\n";
		str += "    end loop;\r\n";
		str += "    return var1;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";
		str += "function std_logic_vector2integer(ARG : std_logic_vector ) return integer is\r\n";
		str += "    variable var1 : unsigned(ARG'length-1 downto 0);\r\n";
		str += "    variable var2 : integer := 0;\r\n";
		str += "begin\r\n";
		str += "    var1 := std_logic_vector2unsigned(ARG);\r\n";
		str += "    var2 := TO_INTEGER(var1);\r\n";
		str += "    return var2;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";
		str += "function std_logic_vector2natural(ARG : std_logic_vector ) return natural is\r\n";
		str += "    variable var1 : unsigned(ARG'length-1 downto 0);\r\n";
		str += "    variable var2 : natural := 0;\r\n";
		str += "begin\r\n";
		str += "    var1 := std_logic_vector2unsigned(ARG);\r\n";
		str += "    var2 := TO_INTEGER(var1);\r\n";
		str += "    return var2;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";
		str += "function std_logic_vector2bit_vector(ARG : std_logic_vector ) return bit_vector is\r\n";
		str += "    variable var1 : bit_vector(ARG'range);\r\n";
		str += "begin\r\n";
		str += "    for I in ARG'range loop\r\n";
		str += "        var1(I) := std_logic2bit(ARG(I));\r\n";
		str += "    end loop;\r\n";
		str += "    return var1;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";
		str += "function std_logic_vector2std_ulogic_vector(ARG : std_logic_vector ) return std_ulogic_vector is\r\n";
		str += "    variable var1 : std_ulogic_vector(ARG'range);\r\n";
		str += "begin\r\n";
		str += "    for I in ARG'range loop\r\n";
		str += "        var1(I) := std_logic2std_ulogic(ARG(I));\r\n";
		str += "    end loop;\r\n";
		str += "    return var1;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";

		str += "function std_logic2bit(ARG : std_logic ) return bit is\r\n";
		str += "    variable var : bit;\r\n";
		str += "begin\r\n";
		str += "    if ( '1' = ARG ) then\r\n";
		str += "        var := '1';\r\n";
		str += "    else\r\n";
		str += "        var := '0';\r\n";
		str += "    end if;\r\n";
		str += "    return var;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";
		str += "function std_logic2std_ulogic(ARG : std_logic ) return std_ulogic is\r\n";
		str += "    variable var : std_ulogic;\r\n";
		str += "begin\r\n";
		str += "    var := std_ulogic(ARG);\r\n";
		str += "    return var;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";
		str += "function std_logic2boolean(ARG : std_logic ) return boolean is\r\n";
		str += "    variable var : boolean;\r\n";
		str += "begin\r\n";
		str += "    if ( '1' = ARG ) then\r\n";
		str += "        var := true;\r\n";
		str += "    else\r\n";
		str += "        var := false;\r\n";
		str += "    end if;\r\n";
		str += "    return var;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";
		str += "function std_logic2character(ARG : std_logic ) return character is\r\n";
		str += "    variable var : character;\r\n";
		str += "begin\r\n";
		str += "    if ( '1' = ARG ) then\r\n";
		str += "        var := '1';\r\n";
		str += "    elsif ( '0' = ARG ) then\r\n";
		str += "        var := '0';\r\n";
		str += "    elsif ( 'X' = ARG ) then\r\n";
		str += "        var := 'x';\r\n";
		str += "    elsif ( '-' = ARG ) then\r\n";
		str += "        var := '-';\r\n";
		str += "    else\r\n";
		str += "        var := 'z';\r\n";
		str += "    end if;\r\n";
		str += "    return var;\r\n";
		str += "end;\r\n";
		str += "\r\n\r\n";
		str += "end package body TB_conv_pack;\r\n";
		str += "\r\n\r\n";

		return str;
	}

	String printHeader() {
		String str = new String();

		str += "--------------------------------------------------------------\r\n";
		str += "--                                                          --\r\n";
		str += "-- This testbench has been generated by the VHDL testbench  --\r\n";
		str += "-- generator named GenTBVHDL.                               --\r\n";
		str += "-- Copyright (C) 2012  edautils.com.                        --\r\n";
		str += "-- Contact: help@edautils.com  for support/info.            --\r\n";
		str += "--                                                          --\r\n";
		str += "--                                                          --\r\n";
		str += "--------------------------------------------------------------\r\n";
		str += "--\r\n";
		str += "--\r\n";
		Date now = new Date();
		String username = System.getProperty("user.name");
		str += "-- Generated by : " + username + " on "
				+ DateFormat.getInstance().format(now) + "\r\n";
		str += "--\r\n";
		str += "--\r\n";
		str += "--------------------------------------------------------------\r\n";
		str += "\r\n";
		return str;
	}
}
