/**
 * 
 */
package com.eu.miscedautils.mygentbvhdl;

import org.antlr.runtime.*;

import com.eu.miscedautils.CommonUtils.LogMessages;
import com.eu.miscedautils.vhdlparser.*;
import com.eu.miscedautils.vhdlparser.VhdlRTLParser.parseVhdlRTL_return;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Iterator;
import java.util.LinkedList;


public class GenerateVHDLTestBench {
	public static void main(String[] args) throws RecognitionException {
		String top = null;
		boolean compile = false;
		String infile = null;
		String worklib = null;
		boolean printSrc = false;
		boolean isSTDPack = false;
		String outfile = "tb.vhd";
		LogMessages.instance().setToolprefix(new String("GenTBVHDL"));

		Integer index = 0;
		String opts = new String(" ");
		printBanner();
		while ((null != args) && (index < args.length)) {
			if (args[index].equals("-in")) {
				index++;
				infile = new String(args[index]);
				opts += "-in " + infile + " ";
			} else if (args[index].equals("-out")) {
				index++;
				outfile = new String(args[index]);
				opts += "-out " + outfile + " ";
			} else if (args[index].equals("-top")) {
				index++;
				top = new String(args[index]);
				opts += "-top " + top + " ";
			} else if (args[index].equals("-compile")) {
				compile = true;
				opts += "-compile " + " ";
			} else if (args[index].equals("-help")) {
				printHelp();
				return;
			} else {
				printHelp();
				return;
			}
			index++;
		}
		if (0 == args.length) {
			printHelp();
			return;
		}
		else if (true == compile && null == infile) {
			LogMessages.instance().print( 115, LogMessages.MSG_SEVERITY.ERROR,  "You have used -compile switch but did not provide the input file ...");
			printHelp();
			return;
		}
		
		FVhPDescription.instance().readPrecompiledLibraryMapFile();
		File mylib = new File("myvhdlwork");
		if ( false == mylib.exists() )
		{
			mylib.mkdir();
			mylib.deleteOnExit();
		}
		if ( null == worklib )
		{
			worklib = "work";
			FVhPDescription.instance().adddLibraryMap(worklib, mylib.toString());
		}
		if ( compile )
		{
			compileFile( worklib, infile, printSrc, isSTDPack );
		}
		else
		{
			if ( null != infile )
			{
				compileFile( worklib, infile, printSrc, isSTDPack );
			}
			generateTestBench( worklib, top, outfile );
		}

		printCompletion(true);
		return; 
	}

	public static Integer compileFile( String worklib, String fName, boolean printSrc, boolean isSTDPack ) throws RecognitionException
	{
		if ( null == worklib )
		{
			worklib = "work";
			LogMessages.instance().print( 116, LogMessages.MSG_SEVERITY.WARNING,  "Working library is not specified, using '-work' as work lib ..." );
		}
		FVhPDescription.setFile(fName);
		ANTLRNoCaseFileStream istream = null;
		try {
			istream = new ANTLRNoCaseFileStream(fName);
		} catch (IOException e) {
			LogMessages.instance().print( 117, LogMessages.MSG_SEVERITY.ERROR,  "File " + fName + " not found ");
			return 1;
		}
		
		VhdlRTLLexer lexer = new VhdlRTLLexer(istream);
		TokenStream tokenStream = new CommonTokenStream(lexer);
		VhdlRTLParser parser = new VhdlRTLParser(tokenStream);
		LogMessages.instance().print( 118, LogMessages.MSG_SEVERITY.NOTE,  "Parsing file " + fName + " ...");
		parseVhdlRTL_return parser_ret = parser.parseVhdlRTL(worklib, isSTDPack);
		if ( printSrc )
		{
			LogMessages.instance().print( 0, LogMessages.MSG_SEVERITY.NONE, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
			String str2 = tokenStream.toString();
			LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, str2);
			LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX");
			String str = FVhPDescription.instance().getRoot().print();
			LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, str);
		}
		return 0;
	}

	public static Integer generateTestBench( String worklib, String duname, String outfile ) throws RecognitionException
	{
		if ( null == worklib )
		{
			worklib = "work";
			LogMessages.instance().print( 119, LogMessages.MSG_SEVERITY.WARNING,  "Working library is not specified, using 'work' as work lib ..." );
		}
		if ( null == duname )
		{
			LogMessages.instance().print( 120, LogMessages.MSG_SEVERITY.ERROR,  
                    "Design unit is not specified to read ...");
			return 1;
		}
		
		String str = null;
		FVhPBase du = null;
		try 
		{
			du = FVhPDescription.instance().restoreDesignUnit( worklib, duname);
			str = du.print();
		} 
		catch (IOException e) 
		{
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		FVhPEntity entity = (FVhPEntity)du;
		entity.elaborate();
		FVhPGenTB tb = new FVhPGenTB(entity, outfile, null, null);
		LogMessages.instance().print( 120, LogMessages.MSG_SEVERITY.NOTE,  "Writing testbench file " + outfile
				+ "...");
		tb.printTB();
		LogMessages.instance().print( 121, LogMessages.MSG_SEVERITY.NOTE,  "Completed testbench ... ");
		
		return 0;
	}

	public static void printHelp() {
		LogMessages.instance().print( 0, LogMessages.MSG_SEVERITY.NOTE, "\nGenTBVHDL: Note: Usage:  -compile [-work libname] -in <input VHDL file> \n");
	    LogMessages.instance().print( 0, LogMessages.MSG_SEVERITY.NONE, "                                       OR \n");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "                           -top <top_entity_name which already compiled or in the given input VHDL file > -in <VHDL file containiting top> [-work libname]\n");
		return;
	}

	public static void printBanner() {
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "                  VHDL TB Generator for Beginners ");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "                Free utility from edautils.com ");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "            Contact info: help@edautils.com");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "     Copyright (C) 2012  Kanai L. Ghosh");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "     ");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "     ");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "     This program is distributed in the hope that it will be useful,");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "     but WITHOUT ANY WARRANTY; without even the implied warranty of");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  ");
		LogMessages.instance().print(0, LogMessages.MSG_SEVERITY.NONE, "     ");

		return;
	}

	public static void printCompletion(boolean successful) {
		if ( 0 == LogMessages.getTotalErrors()) {
			LogMessages.instance().print(125, LogMessages.MSG_SEVERITY.NOTE,  "Generated the output file successfully");
		}
		LogMessages.instance().print(126, LogMessages.MSG_SEVERITY.NOTE,  "Found " + LogMessages.getTotalErrors() + " error(s) and " + LogMessages.getTotalWarnings() + " warning(s) ...");
		LogMessages.instance().print( 127, LogMessages.MSG_SEVERITY.NOTE, "Thank you for using this utility");
		LogMessages.instance().print( 128,LogMessages.MSG_SEVERITY.NOTE,  "Send your suggestion/feedback to help@edautils.com");
		LogMessages.instance().print( 129, LogMessages.MSG_SEVERITY.NOTE,  "Exiting ...\n");
		return;
	}
}
